#include <iostream>
#include <fstream>
#include <sstream>
#include <cmath>
#include <vector>
#include <cstdlib>
#include <ctime>
#include <stdexcept>
#include <stdint.h>

#define STHROW(msg) { \
	 std::ostringstream err;\
	 err<<msg; \
	 throw std::runtime_error(err.str()); }

using namespace std;

template<typename T>  inline 
T SQR (const T& x) {
	return x*x;
}

struct coord {
	double x; 
	double y;
	coord (double x, double y) : x(x), y(y) { }
};

typedef vector<coord> coordlist;

inline double distance(coord point, coord p0, coord direction) {
	// compute the normal distance from point
	// to the line passing through p0 with direction direction
	// direction must be a unit vector
	coord normal(direction.y, -direction.x);
	return abs((point.x-p0.x)*normal.x + (point.y-p0.y)*normal.y);
}

coordlist coords; // Storage for the points

double maxdist=2.0; // maximum distance from the arms of the cross
double armlength=20.0; // length of the arms

/*************************
 * This objective function does not
 * do what is intended
 * ***********************/
double fitnessfunction(vector<double> pos) {
	const double x0=pos[0];
	const double y0=pos[1];
	const double phi=pos[2];

	// PSO maximizes. In order to minize the distance,
	// we return -distance from (x0,y0) -> coords[0]
	//
	return -(sqrt(SQR(coords[0].x-x0)+SQR(coords[0].y-y0)));
}      



/*************************************************************
 * YOU SHOULD NOT NEED TO CHANGE ANYTHING BELOW THIS LINE 
 * **********************************************************/

// This is for including the solution of the homework
// #include "solution.hpp"

// pseudo random number generator MMIX from D. Knuth
static uint64_t randomstate;

double drand() {
	// random number between 0 and 1
	randomstate = randomstate*6364136223846793005ul + 1442695040888963407ul;
	return double(randomstate)/double(0xFFFFFFFFFFFFFFFFul);
}

void drand_seed(uint64_t seed) {
	randomstate=seed;
	drand();
}

// data storage for PSO
struct partmemory {
	// Best seen position and fitness
	// at a specific point
	vector <double> bestpos;
	double fitness;
	// current position and velocity
	vector <double> curpos;
	vector <double> velocity;
	partmemory(int N) : bestpos(N), curpos(N), velocity(N) 
	{ }
};

double pso(vector<double> xmin, vector<double> xmax, vector<double>& result) {
	// Off - the - shelf particle swarm optimization
	// Parameters taken from http://antho.huntingdon.edu/publications/Off-The-Shelf_PSO.pdf
	double phi1=2.8;
	double phi2=1.3;
	double phi=phi1+phi2;
	double K = 2.0/abs(2-phi-sqrt(phi*phi-4*phi));

	const int N=xmin.size();
	// Nr. of dimensions of the problem
	const int pnr=30;

	const int MAXITER=200; 
	// memory of one bird
	vector<partmemory> memory;
	// global memory - best seen position and fitness
	double bestglobalfitness;
	vector<double> bestglobalposition;

	// initialize - generate random bird positions
	for (int bird=0; bird<pnr; bird++) {

		partmemory mybird(N);

		for (int i=0; i<N; i++) {
			double mix=drand();
			mybird.curpos[i]=mix*xmin[i]+(1-mix)*xmax[i];
			mybird.bestpos[i]=mybird.curpos[i];
			mybird.velocity[i]=0;
		}
		mybird.fitness=fitnessfunction(mybird.curpos);
		memory.push_back(mybird);
	}  

	// initialize global best positions
	bestglobalfitness=memory[0].fitness;
	bestglobalposition=memory[0].bestpos;
	for (int bird=1; bird<pnr; bird++) {
		if (memory[bird].fitness>bestglobalfitness) {
			bestglobalfitness=memory[bird].fitness;
			bestglobalposition=memory[bird].bestpos;
		}  
	}

	// now do some iterations

	for (int iter=0; iter<MAXITER; iter++) {
		for (int bird=0; bird<pnr; bird++) {
			// make aliases
			vector<double>& v = memory[bird].velocity;
			vector<double>& x = memory[bird].curpos;
			vector<double>& p = memory[bird].bestpos;
			vector<double>& pg= bestglobalposition;

			// update velocity
			for (int i=0; i<N; i++) {
				double rand1=drand();
				double rand2=drand();

				v[i]=K*(v[i]+phi1*rand1*(p[i]-x[i])+phi2*rand2*(pg[i]-x[i]));
				x[i]=x[i]+v[i];

				if (x[i]>xmax[i]) {
					v[i]=-v[i];
					x[i]=xmax[i];
				}

				if (x[i]<xmin[i]) {
					v[i]=-v[i];
					x[i]=xmin[i];
				}
			}	
			// compute new fitnessfunction fo this bird
			// and, in case it's better here, update individual
			// and global memory
			double newfitness=fitnessfunction(x);
			if (newfitness>memory[bird].fitness) {
				memory[bird].fitness=newfitness;
				p=x;
			}

			if (newfitness>bestglobalfitness) {
				bestglobalfitness=newfitness;
				pg=x;
			}

		}

		cerr<<".";
		if (iter %20 ==0) {
			cerr<<iter<<" Best global fitness: "<<bestglobalfitness<<endl;
		}
	}  
	cerr<<endl;
	result=bestglobalposition;
	return bestglobalfitness;
}

// function to read an ASCII data file
coordlist read_file(const char *fname) {
	ifstream df(fname);
	if (!df) STHROW("Can't read datafile "<<fname);
	coordlist result;
	while (true) {
		string line;
		getline(df, line);
		if (df.eof()) break;
		istringstream stream(line);
		double x, y;
		stream>>x>>y;
		if (!stream) {
			cerr<<"Skipped line "<<line<<endl;
			continue;
		}
		result.push_back(coord(x, y));
	}

	return result;
}

int main(int argc, char **argv) {
	drand_seed(time(NULL));
	try {
		if (argc!=2) STHROW("kreuzfit_pso <datafile.dat>");

		coords = read_file(argv[1]);

		vector<double> xmin(3);
		vector<double> xmax(3);

		// Box constraints for centerpoint
		xmin[0]=0;      xmin[1]=0;
		xmax[0]=300;	xmax[1]=xmax[0];
		// Box constraints for angle
		xmin[2]=-0.2;      xmax[2]=1.6;
		
		// run fit
		vector<double> result(3);
		double bestfit=pso(xmin, xmax, result);
		// get result
		const double x0=result[0];
		const double y0=result[1];
		const double phi=result[2];
		// output for human
		cerr<<x0<<"\t"<<y0<<"\t"
			<<phi*180/M_PI<<"°\t"
			<<bestfit<<endl;

		// output result as a gnu-plottable file
		const double dx = armlength*cos(phi);
		const double dy = armlength*sin(phi);
		cout<<x0+dx<<"\t"<<y0+dy<<endl;
		cout<<x0<<"\t"<<y0<<endl;
		cout<<x0-dx<<"\t"<<y0-dy<<endl;
		cout<<endl;
		cout<<x0+dy<<"\t"<<y0-dx<<endl;
		cout<<x0-dy<<"\t"<<y0+dx<<endl;

	} catch (exception& msg) {
		cerr<<msg.what()<<endl;
		return -1;
	}

	return 0;
}

